const webpack = require('webpack');
const fs = require('fs-extra');
const path = require('path');
const chalk = require('chalk');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');
const log = require('../utils/log');
const Context = require('../core/Context');

module.exports = async function({
  args,
  rootDir,
}) {
  const command = 'build';

  const context = new Context({
    args,
    command,
    rootDir,
  });

  log.verbose(`${command} cliOptions: ${JSON.stringify(args, null, 2)}`);

  const { applyHook, rootDir: ctxRoot } = context;

  await applyHook(`before.${command}.load`, { args });

  let configArr = [];
  try {
    configArr = await context.setUp();
  } catch (err) {
    log.error(chalk.red('Failed to get config.'));
    await applyHook(`error`, { err });
    throw err;
  }

  if (!configArr.length) {
    const err = new Error('No webpack config found.');
    await applyHook(`error`, { err });
    throw err;
  }

  // clear build directory
  const defaultPath = path.resolve(ctxRoot, 'build');  
  configArr.forEach(config => {
    try {
      const userBuildPath = config.output.get('path');
      const buildPath = path.resolve(ctxRoot, userBuildPath);
      fs.removeSync(buildPath);
    } catch (e) {
      if (fs.existsSync(defaultPath)) {
        fs.removeSync(defaultPath);
      }
    }
  });
  
  await applyHook(`before.${command}.run`);
  
  const webpackConfig = configArr.map(v => v.chainConfig.toConfig());
  
  let compiler;
  try {
    compiler = webpack(webpackConfig);
  } catch (err) {
    log.error(chalk.red('Failed to load webpack config.'));
    await applyHook(`error`, { err });
    throw err;
  }

  const result = await new Promise((resolve, reject) => {
    compiler.run((err, stats) => {
      if (err) {
        log.error(err.stack || err);
        if (err.details) {
          log.error(err.details);
        }
        reject(err);
        return;
      }

      log.info(
        stats.toString({
          assets: true,
          colors: true,
          chunks: false,
          entrypoints: false,
          modules: false,
        }),
      );

      const json = stats.toJson({}, true);
      const messages = formatWebpackMessages(json);
      const isSuccessful = !messages.errors.length && !messages.warnings.length;

      if (isSuccessful) {
        if (stats.stats) {
          log.info('Compiled successfully');
        } else {
          log.info(
            `Compiled successfully in ${(json.time / 1000).toFixed(1)}s!`
          );
        }
      } else if (messages.errors.length) {
        console.log(messages.errors.join('\n\n'));
      } else if (messages.warnings.length) {
        log.warn('Compiled with warnings.');
        console.log(messages.warnings.join('\n\n'));
      }

      resolve({
        stats,
      });
    });
  });

  await applyHook(`after.${command}.compile`, result);
};
