const webpack = require('webpack');
const chalk = require('chalk');
const deepmerge = require('deepmerge');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');
const WebpackDevServer = require('webpack-dev-server');
const prepareUrLs = require('../utils/prepareURLs');
const log = require('../utils/log');
const Context = require('../core/Context');

module.exports = async function({
  args,
  rootDir,
}) {
  const command = 'start';

  const context = new Context({
    args,
    command,
    rootDir,
  });

  log.verbose(`${command} cliOptions: ${JSON.stringify(args, null, 2)}`);
  let serverUrl = '';

  const { applyHook } = context;

  await applyHook(`before.${command}.load`, { args });

  let configArr = [];
  try {
    configArr = await context.setUp();
  } catch (err) {
    log.error(chalk.red('Failed to get config.'));
    await applyHook(`error`, { err });
    throw err;
  }

  if (!configArr.length) {
    const err = new Error('No webpack config found.');
    await applyHook(`error`, { err });
    throw err;
  }

  let devServerConfig = {
    port: args.port || 9999,
    host: args.host || '0.0.0.0',
  };

  for (const item of configArr) {
    const { chainConfig } = item;

    const webpackConfig = chainConfig.toConfig();

    if (webpackConfig.devServer) {
      devServerConfig = deepmerge(devServerConfig, webpackConfig.devServer);
    }
  }

  await applyHook(`before.${command}.run`);

  const webpackConfig = configArr.map(v => v.chainConfig.toConfig());

  let compiler;
  try {
    compiler = webpack(webpackConfig);
  } catch (err) {
    log.error(chalk.red('Failed to load webpack config.'));
    await applyHook(`error`, { err });
    throw err;
  }
  const protocol = devServerConfig.https ? 'https' : 'http';
  const urls = prepareUrLs(protocol, devServerConfig.host, devServerConfig.port);
  serverUrl = urls.localUrlForBrowser;

  compiler.hooks.done.tap('compileHook', async (stats) => {
    log.info(
      stats.toString({
        assets: true,
        colors: true,
        chunks: false,
        entrypoints: false,
        modules: false,
      }),
    );

    const json = stats.toJson({}, true);
    const messages = formatWebpackMessages(json);
    const isSuccessful = !messages.errors.length && !messages.warnings.length;

    if (isSuccessful) {
      if (stats.stats) {
        log.info('Compiled successfully');
      } else {
        log.info(
          `Compiled successfully in ${(json.time / 1000).toFixed(1)}s!`
        );
      }
    } else if (messages.errors.length) {
      console.log(messages.errors.join('\n\n'));
    } else if (messages.warnings.length) {
      log.warn('Compiled with warnings.');
      console.log(messages.warnings.join('\n\n'));
    }

    await applyHook(`after.${command}.compile`, {
      url: serverUrl,
      stats,
    });
  });

  const devServer = new WebpackDevServer(compiler, devServerConfig);

  await applyHook(`before.${command}.devServer`, {
    url: serverUrl,
    devServer,
  });

  devServer.listen(devServerConfig.port, devServerConfig.host, async(err) => {
    if (err) {
      log.info(chalk.red('[ERR]: Failed to start webpack dev server'));
      log.error(err.message || err);
    }

    console.log();
    log.info(chalk.green('Starting the development server at:'));
    log.info('   ', chalk.underline.white(serverUrl));
    console.log();

    await applyHook(`after.${command}.devServer`, {
      url: serverUrl,
      devServer,
      err,
    });
  });

  return devServer;
};
