const path = require('path');
const assert = require('assert');
const fs = require('fs-extra');
const _ = require('lodash');
const camelCase = require('camelcase');
const webpack = require('webpack');
const log = require('../utils/log');

const PKG_FILE = 'package.json';
const USER_CONFIG_FILE = 'build.json';
const PLUGIN_CONTEXT_KEY = [
  'command',
  'commandArgs',
  'rootDir',
  'userConfig',
  'pkg',
  'webpack',
];

const VALIDATION_MAP = {
  string: 'isString',
  number: 'isNumber',
  array: 'isArray',
  object: 'isObject',
  boolean: 'isBoolean',
};

const BUILDIN_CLI_OPTIONS = [
  { name: 'port', commands: ['start'] },
  { name: 'host', commands: ['start'] },
  { name: 'config', commands: ['start', 'build', 'test'] },
];

module.exports = class Context {
  constructor({
    command,
    rootDir = process.cwd(),
    args = {} }
  ) {
    // bind api
    [
      'getProjectFile', 'getUserConfig', 'getPlugins', 'getAllTask',
      'onHook', 'applyHook',
      'setValue', 'getValue',
      'registerTask', 'registerUserConfig', 'registerCliOption',
      'runPlugins', 'runUserConfig', 'runWebpackFunctions', 'runCliOption',
      'onGetWebpackConfig',
      'setUp',
    ].forEach((api) => {
      this[api] = this[api].bind(this);
    });

    this.command = command;
    this.commandArgs = args;
    this.rootDir = rootDir;
    this.webpack = webpack;
    /**
     * config array
     * {
     *   name,
     *   chainConfig,
     *   webpackFunctions,
     * }
     */
    this.configArr = [];
    this.modifyConfigFns = [];
    this.eventHooks = {}; // lifecycle functions
    this.internalValue = {}; // internal value shared between plugins
    this.userConfigRegistration = {};
    this.cliOptionRegistration = {};

    this.pkg = this.getProjectFile(PKG_FILE);
    this.userConfig = this.getUserConfig();
    this.checkPluginValue(); // check plugins property
    // register buildin options
    this.registerCliOption(BUILDIN_CLI_OPTIONS);
    this.plugins = this.getPlugins();
  }

  _registerConfig(type, args, parseName) {
    const registerKey = `${type}Registration`;
    if (!this[registerKey]) {
      throw new Error(`unknown register type: ${type}, use available types (userConfig or cliOption) instead`);
    }
    const configArr = _.isArray(args) ? args : [args];
    configArr.forEach((conf) => {
      const confName = parseName ? parseName(conf.name) : conf.name;
      if (this[registerKey][confName]) {
        throw new Error(`${conf.name} already registered in ${type}`);
      }

      this[registerKey][confName] = conf;

      // set default userConfig
      if (type === 'userConfig'
        && _.isUndefined(this.userConfig[confName])
        && Object.prototype.hasOwnProperty.call(conf, 'defaultValue')) {
        this.userConfig[confName] = conf.defaultValue;
      }
    });
  }

  async _runConfigWebpack(fn, configValue) {
    for (const webpackConfigInfo of this.configArr) {
      const userConfigContext = _.pick(this, PLUGIN_CONTEXT_KEY);
      userConfigContext.taskName = webpackConfigInfo.name;

      await fn(webpackConfigInfo.chainConfig, configValue, userConfigContext);
    }
  }

  getProjectFile(fileName) {
    const configPath = path.resolve(this.rootDir, fileName);

    let config = {};
    if (fs.existsSync(configPath)) {
      try {
        config = fs.readJsonSync(configPath);
      } catch (err) {
        log.info('CONFIG', `Fail to load config file ${configPath}, use empty object`);
      }
    }

    return config;
  }

  getUserConfig() {
    const { config } = this.commandArgs;
    let configPath = '';
    if (config) {
      configPath = path.isAbsolute(config) ? config : path.resolve(this.rootDir, config);
    } else {
      configPath = path.resolve(this.rootDir, USER_CONFIG_FILE);
    }
    let userConfig = {};
    if (fs.existsSync(configPath)) {
      try {
        userConfig = fs.readJsonSync(configPath); // read build.json
      } catch (err) {
        log.info('CONFIG', `Fail to load config file ${configPath}, use default config instead`);
        log.error(err);
        process.exit(1);
      }
    }

    return userConfig;
  }

  getPlugins() {
    const userPlugins = this.userConfig.plugins.map((pluginInfo) => {
      let fn = () => {};

      const plugins = Array.isArray(pluginInfo) ? pluginInfo : [pluginInfo];

      const pluginPath = require.resolve(plugins[0], { paths: [this.rootDir] });
      const options = plugins[1];

      try {
        fn = require(pluginPath); // eslint-disable-line
      } catch (err) {
        log.error(`Fail to load plugin ${pluginPath}`);
        log.error(err);
        process.exit(1);
      }

      return {
        name: plugins[0],
        fn,
        options,
      };
    });

    return userPlugins;
  }

  registerTask(name, chainConfig) {
    const exist = this.configArr.find(v => v.name === name);
    if (!exist) {
      this.configArr.push({
        name,
        chainConfig,
        modifyFunctions: [],
      });
    } else {
      throw new Error(`[Error] config '${name}' already exists!`);
    }
  }

  getAllTask() {
    return this.configArr.map(v => v.name);
  }

  onGetWebpackConfig(name, fn) {
    this.modifyConfigFns.push([name, fn]);
  }

  onHook(key, fn) {
    if (!Array.isArray(this.eventHooks[key])) {
      this.eventHooks[key] = [];
    }
    this.eventHooks[key].push(fn);
  }

  async applyHook(key, opts = {}) {
    const hooks = this.eventHooks[key] || [];

    for (const fn of hooks) {
      await fn(opts);
    }
  }

  setValue(key, value) {
    this.internalValue[key] = value;
  }

  getValue(key) {
    return this.internalValue[key];
  }

  registerUserConfig(args) {
    this._registerConfig('userConfig', args);
  }

  registerCliOption(args) {
    this._registerConfig('cliOption', args, (name) => {
      return camelCase(name, { pascalCase: false });
    });
  }

  async runPlugins() {
    for (const pluginInfo of this.plugins) {
      const { fn, options } = pluginInfo;

      const pluginContext = _.pick(this, PLUGIN_CONTEXT_KEY);

      const pluginAPI = {
        log,
        context: pluginContext,
        registerTask: this.registerTask,
        getAllTask: this.getAllTask,
        onGetWebpackConfig: this.onGetWebpackConfig,
        onHook: this.onHook,
        setValue: this.setValue,
        getValue: this.getValue,
        registerUserConfig: this.registerUserConfig,
        registerCliOption: this.registerCliOption,
      };

      await fn(pluginAPI, options);
    }
  }

  checkPluginValue() {
    const value = this.userConfig.plugins;
    let flag;
    if(!_.isArray(value)) {
      flag = false;
    } else {
      flag = value.every(v => {
        let correct = _.isArray(v) || _.isString(v);
        if (correct && _.isArray(v)) {
          correct = _.isString(v[0]);
        }

        return correct;
      });
    }

    if(!flag) {
      throw new Error(`plugins did not pass validation, result: ${flag}`);
    }
  }

  async runUserConfig() {
    for (const configInfoKey in this.userConfig) {
      if (configInfoKey === 'plugins') continue;

      const configInfo = this.userConfigRegistration[configInfoKey];

      if (!configInfo) {
        throw new Error(`[build.json] Config key '${configInfoKey}' is not supported`);
      }

      const { name, validation } = configInfo;
      const configValue = this.userConfig[name];

      if (validation) {
        let validationInfo;
        if (_.isString(validation)) {
          const fnName = VALIDATION_MAP[validation];
          if (!fnName) {
            throw new Error(`validation does not support ${validation}`);
          }
          assert(_[VALIDATION_MAP[validation]](configValue), `Config ${name} should be ${validation}, but got ${configValue}`);
        } else {
          validationInfo = await validation(configValue);
          assert(validationInfo, `${name} did not pass validation, result: ${validationInfo}`);
        }
      }

      if (configInfo.configWebpack) {
        await this._runConfigWebpack(configInfo.configWebpack, configValue);
      }
    }
  }

  async runCliOption() {
    for (const cliOpt in this.commandArgs) {
      // allow all jest option when run command test
      if (this.command === 'test' && cliOpt === 'jestArgv') {
        continue;
      }
      const { commands, name, configWebpack } = this.cliOptionRegistration[cliOpt] || {};
      if (!name || !(commands || []).includes(this.command)) {
        throw new Error(`cli option '${cliOpt}' is not supported when run command '${this.command}'`);
      }
      if (configWebpack) {
        await this._runConfigWebpack(configWebpack, this.commandArgs[cliOpt]);
      }
    }
  }

  async runWebpackFunctions() {
    this.modifyConfigFns.forEach(([name, func]) => {
      const isAll = _.isFunction(name);
      if (isAll) {  // modify all
        this.configArr.forEach(config => {
          config.modifyFunctions.push(name);
        });
      } else { // modify named config
        this.configArr.forEach(config => {
          if (config.name === name) {
            config.modifyFunctions.push(func);
          }
        });
      }
    });
    
    for (const configInfo of this.configArr) {
      for (const func of configInfo.modifyFunctions) {
        await func(configInfo.chainConfig);
      }
    }
  }

  async setUp() {
    await this.runPlugins();
    await this.runUserConfig();
    await this.runWebpackFunctions();
    await this.runCliOption();

    return this.configArr;
  }
};
